// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package pathfinder

import (
	"fmt"
	"github.com/google/gofuzz"
	"github.com/stretchr/testify/assert"
	"testing"
)

type Issue2SubSubStruct struct {
	Issue2SubSubField []string
}

type Issue2SubStruct struct {
	Issue2SubField Issue2SubSubStruct
}

type Issue2Struct struct {
	Issue1Sub1 Issue2SubStruct
}

func TestGetValueWithArray(t *testing.T) {
	testStructForGet := Issue2Struct{
		Issue1Sub1: Issue2SubStruct{
			Issue2SubField: Issue2SubSubStruct{
				Issue2SubSubField: []string{"test0", "test1", "test2"},
			},
		},
	}

	// iterate from 0 to 2
	for i := 0; i < 2; i++ {
		result, err := GetValue(testStructForGet, "Issue1Sub1.Issue2SubField.Issue2SubSubField."+fmt.Sprintf("%d", i))

		assert.Nil(t, err)
		assert.Equal(t, result, "test"+fmt.Sprintf("%d", i))
	}

	i := 3
	result, err := GetValue(testStructForGet, "Issue1Sub1.Issue2SubField.Issue2SubSubField."+fmt.Sprintf("%d", i))

	assert.NotNil(t, err)
	assert.Nil(t, result)

}

func TestGetValueWithArrayFuzz(t *testing.T) {
	f := fuzz.New()

	testStructForGet := Issue2Struct{
		Issue1Sub1: Issue2SubStruct{
			Issue2SubField: Issue2SubSubStruct{
				Issue2SubSubField: []string{"test0", "test1", "test2"},
			},
		},
	}

	for i := 0; i < 100; i++ {
		var randomIndex int
		f.Fuzz(&randomIndex)
		randomIndex = randomIndex % len(testStructForGet.Issue1Sub1.Issue2SubField.Issue2SubSubField)
		if randomIndex < 0 {
			randomIndex = -randomIndex
		}

		result, err := GetValue(testStructForGet, "Issue1Sub1.Issue2SubField.Issue2SubSubField."+fmt.Sprintf("%d", randomIndex))

		if randomIndex < 3 {
			assert.Nil(t, err)
			assert.Equal(t, result, "test"+fmt.Sprintf("%d", randomIndex))
		} else {
			assert.NotNil(t, err)
			assert.Nil(t, result)
		}
	}
}

func TestSetValueWithArray(t *testing.T) {
	testStructForSet := Issue2Struct{
		Issue1Sub1: Issue2SubStruct{
			Issue2SubField: Issue2SubSubStruct{
				Issue2SubSubField: []string{"test0", "test1", "test2"},
			},
		},
	}

	// iterate from 0 to 2
	for i := 0; i < 2; i++ {

		newValue := "test~~" + fmt.Sprintf("%d", i)
		k := "Issue1Sub1.Issue2SubField.Issue2SubSubField." + fmt.Sprintf("%d", i)
		err := SetValue(&testStructForSet, k, newValue)

		assert.Nil(t, err)

		result, err := GetValue(testStructForSet, k)
		assert.Equal(t, result, newValue)
	}

	i := 3
	k := "Issue1Sub1.Issue2SubField.Issue2SubSubField." + fmt.Sprintf("%d", i)
	err := SetValue(testStructForSet, k, "test3")
	assert.NotNil(t, err)

}

type PathValue string

type SubTestSubPaths struct {
	Template    PathValue
	Definitions []PathValue
}

type SubTest2Def struct {
	Paths SubTestSubPaths
}

type SubTestStruct1 map[string]SubTest2Def
type MainTestStruct struct {
	Sub SubTestStruct1
}

func TestReplacePathForConfig(t *testing.T) {
	config := MainTestStruct{
		Sub: SubTestStruct1{
			"Default": SubTest2Def{
				Paths: SubTestSubPaths{
					Template:    "../../../default.html",
					Definitions: []PathValue{"../../../legacy.yaml"},
				},
			},
		},
	}

	nn, err2 := GetValue[MainTestStruct](config, "Sub.Default.Paths.Template")
	assert.Nil(t, err2)

	assert.Equal(t, nn, PathValue("../../../default.html"))

	err := SetValue[*MainTestStruct](&config, "Sub.Default.Paths.Template", "test")
	assert.Nil(t, err)

	nn, err3 := GetValue[MainTestStruct](config, "Sub.Default.Paths.Template")
	assert.Nil(t, err3)

	assert.Equal(t, nn, PathValue("test"))

}
